import inspect
from collections import namedtuple
from .compat import getfullargspec


ArgumentInfo = namedtuple("ArgumentInfo", ["default", "type"])


class NoDefault(object):

    def __str__(self):
        return "NoDefault"

    def __repr__(self):
        return "NoDefault"

NoDefault = NoDefault()


class TransmuteFunction(object):
    """
    VoodooFunctions are objects that wrap a method, allowing
    extensions to extract metadata for their own use (such as
    automatic documentation)
    """

    params = ["error_exceptions"]

    def __init__(self, func, error_exceptions=None):
        # arguments should be the arguments passed into
        # the function
        #
        # the return type should be specified. If nothing is
        # passed in, a NoneType is returned.
        # for a dynamic language like Python, it's not a huge deal:
        # transmute will still json serialize whatever object you pass it.
        # however, the return type is valuable for autodocumentation systems.
        self.arguments, self.return_type = _extract_arguments_and_return_type(func)
        # description should be a description of the api
        # endpoint, for use in autodocumentation
        self.description = func.__doc__
        # error_exceptions represents the exceptions
        # that should be caught and return an API exception
        self.error_exceptions = error_exceptions
        # these are set if the function performs some
        # operation that modifies data.
        self.creates = getattr(func, "creates", False)
        self.deletes = getattr(func, "deletes", False)
        self.updates = getattr(func, "updates", False)
        # produces represents the return types supported
        # by the final function
        self.produces = ["application/json"]
        self.raw_func = func
        # status_codes represents the possible status codes
        # the function can return
        self.responses = _get_responses(self.return_type)
        # this is to make discovery easier.
        # TODO: make sure this doesn't mess up GC, as it's
        # a cyclic reference.
        if inspect.ismethod(func):
            func.__func__.transmute_func = self
        else:
            func.transmute_func = self


def _extract_arguments_and_return_type(func):
    """
    return a dict of <name, type> pairs and a <type>,
    which represent arguments and a return type that
    should passed into the function and returned.
    """
    argspec = getfullargspec(func)
    attributes = (getattr(argspec, "args", []) +
                  getattr(argspec, "keywords", []))
    defaults = argspec.defaults or []

    arguments = {}
    for i, name in enumerate(reversed(attributes)):
        if name == "self":
            continue

        default = NoDefault if len(defaults) <= i else defaults[i]
        arguments[name] = ArgumentInfo(
            default, argspec.annotations.get(name, str)
        )

    return arguments, argspec.annotations.get("return", type(None))


def _get_responses(return_type):
    return {
        200: {
            "description": "success",
            "return_type": {"properties": {"success": {"type": bool},
                                           "result": {"type": return_type}},
                            "required": ["success", "result"]}
        },
        400: {
            "description": "invalid input received",
            "return_type": {"properties": {"success": {"type": bool},
                                           "message": {"type": str}},
                            "required": ["success", "message"]}
        }
    }



import flask_transmute


def test_route_decorator():
    """ RouteSets should have a route decorator method. """
    route_set = flask_transmute.FlaskRouteSet()

    @route_set.route("/foo")
    def test():
        return {"foo": "bar"}

    assert len(route_set._routes) == 1
    assert route_set._routes[0].transmute_func.raw_func == test
    assert route_set._routes[0].path == "/foo"



from .class_serializer import (
    class_is_serializable, generate_class_serializer
)
from .list_serializer import (
    list_is_serializable, generate_list_serializer
)
from .basetype_serializers import (
    BoolSerializer,
    NoneSerializer,
    IntSerializer,
    FloatSerializer,
    StringSerializer
)

# the serializer cache contains serializers for
# all known types.
_SERIALIZER_CACHE = {
    bool: BoolSerializer,
    type(None): NoneSerializer,
    None: NoneSerializer,
    int: IntSerializer,
    float: FloatSerializer,
    str: StringSerializer
}


def get_serializer(cls):
    cls = _make_signature_hashable(cls)

    if cls in _SERIALIZER_CACHE:
        return _SERIALIZER_CACHE[cls]

    _assert_type_is_serializable(cls)

    if isinstance(cls, tuple):
        serializer = generate_list_serializer(cls[0], _SERIALIZER_CACHE)
    else:
        serializer = generate_class_serializer(cls, _SERIALIZER_CACHE)

    _SERIALIZER_CACHE[cls] = serializer
    return serializer


def _make_signature_hashable(cls):
    """
    convert the class signature to a hashable object
    """
    if isinstance(cls, list):
        return (cls[0],)
    return cls


def _assert_type_is_serializable(cls):
    if cls in _SERIALIZER_CACHE:
        return

    if isinstance(cls, tuple):
        for subclass in list_is_serializable(cls):
            get_serializer(subclass)

    else:
        for subclass in class_is_serializable(cls):
            get_serializer(subclass)



import inspect


def get_public_callables(obj):
    """
    return an iterator over all public callables of a function
    """
    for key in dir(obj):
        if key.startswith('_'):
            continue
        value = getattr(obj, key)
        if callable(value):
            yield key, value


def get_raw_function(func):
    if inspect.ismethod(func):
        return func.__func__
    else:
        return func


def join_url_paths(*parts):
    path = parts[0]
    for p in parts[1:]:
        if path == "":
            path = p
        else:
            path = "{0}/{1}".format(path.rstrip("/"), p.lstrip("/"))
    return path



from .exceptions import SerializerException


def list_is_serializable(cls):
    yield cls[0]


def generate_list_serializer(cls, serializers):

    sub_serializer = serializers.get(cls)

    class ListSerializer(object):

        @staticmethod
        def serialize(obj):
            return [
                sub_serializer.serialize(e) for e in obj
            ]

        @staticmethod
        def deserialize(data):
            if not isinstance(data, list):
                raise SerializerException(
                    "unable to serialize a {0} object into a list.".format(type(data.__name__))
                )

            return [sub_serializer.deserialize(e) for e in data]

    ListSerializer.__name__ = "{0}ListSerializer".format(
        cls.__name__
    )

    return ListSerializer



import json
import yaml
import functools
from flask import jsonify, request, current_app
from ..serializers import get_serializer, SerializerException
from ..function import NoDefault
from ..exceptions import ApiException


def wrap_method(transmute_function):
    """
    this handles the conversion of a function
    that returns arbitrary functionality into
    """
    tf = transmute_function

    is_post_method = tf.creates or tf.updates
    api_exceptions = tuple(
        list(tf.error_exceptions or []) + [ApiException, SerializerException]
    )
    args_not_empty = len(tf.arguments) > 0
    result_serializer = get_serializer(tf.return_type)

    @functools.wraps(tf.raw_func)
    def wrapper_func(*args, **kwargs):
        try:
            request_params = _retrieve_request_params(args_not_empty, is_post_method)
            _add_request_parameters_to_args(tf.arguments, request_params, kwargs)
            result = tf.raw_func(*args, **kwargs)
            result = result_serializer.serialize(result)
            return _return({
                "success": True,
                "result": result
            })
        except Exception as e:
            if api_exceptions is not None and isinstance(e, api_exceptions):
                return _return({
                    "success": False,
                    "detail": str(e)
                }, status_code=getattr(e, "status_code", 400))
            else:
                raise

    return wrapper_func


def _retrieve_request_params(args_not_empty, is_post_method):
    if not is_post_method:
        request_args = request.args
    else:
        if (not request.content_type or "json" in request.content_type) and request.get_data():
            try:
                request_args = json.loads(request.get_data().decode("UTF-8"))
                if not isinstance(request_args, dict) and args_not_empty:
                    raise ApiException("expected json object: {0}".format(str(e)))
            except ValueError as e:
                raise ApiException("unable to parse json: {0}".format(str(e)))
        else:
            request_args = request.form
    return request_args


def _add_request_parameters_to_args(arguments, request_args, arg_dict):
    for argument, info in arguments.items():
        if argument in arg_dict:
            continue

        if argument not in request_args:
            if info.default is NoDefault:
                raise ApiException("parameter {0} is required".format(argument))
            else:
                continue
        try:
            serializer = get_serializer(info.type)
            if isinstance(info.type, list):
                value = request_args.getlist(argument)
            else:
                value = request_args.get(argument)
            value = serializer.deserialize(value)

        except SerializerException as e:
            raise ApiException("parameter {0}: {1}".format(argument, str(e)))
        arg_dict[argument] = value


def _return(data, status_code=200):
    content_type = request.content_type or "application/json"
    if "yaml" in content_type:
        return current_app.response_class(
            yaml.dump(data, default_flow_style=False),
            mimetype='application/x-yaml'
        ), status_code
    if "json" in content_type:
        return jsonify(data), status_code
    else:
        return jsonify(data), status_code



class SerializerException(Exception):
    pass


class InvalidSchema(SerializerException):

    def __init__(self, message, schema):
        super(SerializerException, self).__init__(message)
        self.schema = schema


class SerializationException(SerializerException):
    pass


class DeserializationException(SerializerException):
    pass



from .exceptions import SerializationException
from ..compat import string_type


class IntSerializer(object):

    @staticmethod
    def serialize(obj):
        try:
            return int(obj)
        except ValueError:
            raise SerializationException(
                "unable to interpret {0} as an int.".format(str(obj)))

    @staticmethod
    def deserialize(data):
        if isinstance(data, int):
            return data

        try:
            return int(data)
        except ValueError:
            raise SerializationException(
                "unable to interpret {0} as an int.".format(str(data)))


class FloatSerializer(object):

    @staticmethod
    def serialize(obj):
        return obj

    @staticmethod
    def deserialize(data):
        if isinstance(data, float):
            return data

        try:
            return float(data)
        except ValueError:
            raise SerializationException(
                "unable to interpret {0} as an float.".format(str(data)))


class BoolSerializer(object):

    @staticmethod
    def serialize(obj):
        return obj

    @staticmethod
    def deserialize(data):
        if isinstance(data, bool):
            return data

        if not isinstance(data, string_type):
            msg = "unable to interpret data of type {0} as a bool"
            raise SerializationException(msg.format(type(data).__name__))

        data = data.lower()
        return data.startswith("t") or data.startswith("y")


class StringSerializer(object):

    @staticmethod
    def serialize(obj):
        return obj

    @staticmethod
    def deserialize(data):
        if not isinstance(data, string_type):
            msg = "unable to interpret data of type {0} as a string"
            raise SerializationException(
                msg.format(type(data).__name__))
        return data


class NoneSerializer(object):

    @staticmethod
    def serialize(obj):
        return obj

    @staticmethod
    def deserialize(data):
        if data is not None:
            raise SerializationException("Unable to deserialize {0} to None".format(
                data
            ))
        return None



import sys
import inspect
from collections import namedtuple

is_py3 = sys.version_info[0] >= 3

FullArgSpec = namedtuple(
    "FullArgSpec",
    ["args", "varargs", "varkw", "defaults",
     "kwonlyargs", "kwonlydefaults", "annotations"]
)


def getfullargspec(func):
    if is_py3:
        return inspect.getfullargspec(func)
    else:
        argspec = inspect.getargspec(func)
        annotations = getattr(func, "__annotations__", {})
        return FullArgSpec(
            argspec.args, argspec.varargs,
            argspec.keywords, argspec.defaults,
            [], None, annotations
        )

if is_py3:
    string_type = str
else:
    string_type = basestring



def updates(f):
    """
    this labels a function as one that updates data.
    """
    f.updates = True
    return f


def creates(f):
    """
    this labels a function as one that creates data.
    """
    f.creates = True
    return f


def deletes(f):
    """
    this labels a function as one that deletes data.
    """
    f.deletes = True
    return f


def annotate(annotations):
    """
    in python2, native annotions on parameters do not exist:
    def foo(a : str, b: int) -> bool:
        ...

    this provides a way to provide attribute annotations:

    @annotate({"a": str, "b": int, "return": bool})
    def foo(a, b):
        ...
    """

    def decorate(func):
        func.__annotations__ = annotations
        return func

    return decorate



from abc import ABCMeta, abstractmethod
from collections import namedtuple
from .function import TransmuteFunction
from .utils import get_public_callables


RouteConfig = namedtuple("RouteConfig", ["path", "transmute_func"])
SubRouteSet = namedtuple("SubRouteSet", ["path", "route_set"])


class RouteSet(object):
    """
    a base class for a set of routes. a particular framework should
    extend this class and override methods as necessary.
    """
    __metaclass__ = ABCMeta

    # this is a list of parameters that are important in the
    # context of transmute. This is available for extended route sets
    # to make decisions about which parameters to capture and which to pass
    # through to the framework's routing mechanism.
    transmute_params = TransmuteFunction.params

    def add_route_set(self, path, route_set):
        self._route_set_pairs.append(SubRouteSet(path, route_set))

    # other methods that may be worth overriding
    def init_app(self, *args, **kwargs):

        for subroute_set in self._route_set_pairs:
            subroute_set.init_app(*args, **kwargs)

        for extension in self._extensions:
            extension.init_app(self, *args, **kwargs)

    # route function
    def route_function(self, path, function, **options):
        transmute_func = TransmuteFunction(function, **options)
        route_config = RouteConfig(path, transmute_func)
        self._routes.append(route_config)

    def route(self, path, **options):
        """
        to accomodate a more flask-like syntax, you can decorate a single
        method
        """

        def decorator(func):
            self.route_function(path, func, **options)
            return func

        return decorator

    @classmethod
    def _split_options_dict(cls, options_dict):
        """
        oftentimes specific route sets will find the need
        to split parameters between what should be consumed by transmute
        and what should be consumed by your framework's routing mechanism.

        this is a convenience method provided that will help that split for you,
        dividing any options dictionary into {transmute_options}, {other_options}.
        """
        transmute_options, other_options = {}, {}

        for k, v in options_dict.items():
            if k in cls.transmute_params:
                transmute_options[k] = v
            else:
                other_options[k] = v

        return transmute_options, other_options

    def __init__(self):
        self._routes = []
        self._extensions = []
        self._route_set_pairs = []

    def route_object(self, path, obj, **options):
        for method_name, func in get_public_callables(obj):
            func_path = "{0}/{1}".format(path, method_name)
            self.route_function(func_path, func, **options)



import types
from .exceptions import (
    SerializerException, SerializationException
)
from .schema import validate_schema


def class_is_serializable(cls):

    from_dict = getattr(cls, "from_transmute_dict", None)

    if not from_dict or not isinstance(from_dict, types.FunctionType):
        raise SerializerException(
            "expected from_transmute_dict for {0} to exist and be a function!".format(cls))

    schema = getattr(cls, "transmute_schema", None)
    validate_schema(schema)
    for detail in schema["properties"].values():
        yield detail["type"]


def generate_class_serializer(cls, serializers):
    schema = cls.transmute_schema
    from_transmute_dict = cls.from_transmute_dict

    class ClassSerializer(object):
        """ serializes an object to and from a markup-serializable type """

        @staticmethod
        def serialize(obj):
            data = {}
            for attr_name, attr_details in schema["properties"].items():
                serializer = serializers[attr_details["type"]]
                if hasattr(obj, attr_name):
                    value = getattr(obj, attr_name)
                    data[attr_name] = serializer.serialize(value)
                elif attr_name in schema.get("required", []):
                    raise SerializationException(
                        "required attribute {0} not found".format(attr_name)
                    )
            return data

        @staticmethod
        def deserialize(data):
            return from_transmute_dict(data)

    ClassSerializer.__name__ = "{0}Serializer".format(
        cls.__name__
    )

    serializers[cls] = ClassSerializer
    return ClassSerializer



import pytest
from flask_transmute.utils import join_url_paths


@pytest.mark.parametrize("inp, expected_output", [
    (["", "foobar"], "foobar")
])
def test_join_url_paths(inp, expected_output):
    assert join_url_paths(*inp) == expected_output



from ..routeset import RouteSet
from .method_wrapper import wrap_method
from ..function import TransmuteFunction
from ..utils import join_url_paths


class FlaskRouteSet(RouteSet):

    def __init__(self, *args, **kwargs):
        self._route_options = {}
        super(FlaskRouteSet, self).__init__(*args, **kwargs)

    def init_app(self, app, path=""):
        for route_config in self._routes:
            _route_transmute_func(app,
                                  join_url_paths(path, route_config.path),
                                  route_config.transmute_func)
        for path, route_set in self._route_set_pairs:
            route_set.init_app(app, path=path)

    def route_function(self, path, function, **options):
        transmute_options, flask_options = self._split_options_dict(options)
        self._route_options[path] = flask_options
        super(FlaskRouteSet, self).route_function(path, function,
                                                  **transmute_options)


def _route_transmute_func(router, path, transmute_func, **options):
    method = "GET"
    if transmute_func.creates:
        method = "PUT"
    elif transmute_func.deletes:
        method = "DELETE"
    elif transmute_func.updates:
        method = "POST"
    return router.route(path, methods=[method], **options)(wrap_method(transmute_func))



from ..function import NoDefault


class Paths(object):

    def __init__(self, definitions):
        self._paths = {}
        # this is a Definitions object.
        self._definitions = definitions

    def extract_from_app(self, app):
        for rule in app.url_map.iter_rules():
            path = rule.rule
            endpoint = rule.endpoint
            func = app.view_functions[endpoint]
            if hasattr(func, "transmute_func"):
                if path not in self._paths:
                    self._paths[path] = {}

                self._paths[path].update(
                    self._extract_swagger_pathspec(func.transmute_func)
                )

    def add_to_spec(self, spec):
        spec["paths"] = self._paths

    def _extract_swagger_pathspec(self, transmute_func):
        path_spec = {
            "description": transmute_func.description,
            "produces": transmute_func.produces,
            "parameters": [],
            "responses": {}
        }

        if transmute_func.updates or transmute_func.creates:
            # in the case of updates or creates, we spec this by
            # specifying all values are passed into the body.
            param_spec = {"name": "body", "in": "body", "required": True}
            properties = {}
            required = []
            for arg_name, arg_info in transmute_func.arguments.items():
                properties[arg_name] = {"type": arg_info.type}
                if arg_info.default is NoDefault:
                    required.append(arg_name)
            model = {"properties": properties, "required": required}
            param_spec["schema"] = self._definitions.get(model)
            path_spec["parameters"].append(param_spec)

        else:
            for arg_name, arg_info in transmute_func.arguments.items():
                param_spec = {
                    "name": arg_name,
                    "required": arg_info.default is not None,
                    "in": "query"
                }
                type_definition = self._definitions.get(arg_info.type)
                param_spec.update(type_definition)

                path_spec["parameters"].append(param_spec)

        for code, details in transmute_func.responses.items():
            response_spec = {"description": details["description"]}
            if details["return_type"] is not type(None):
                response_spec["schema"] = self._definitions.get(details["return_type"])
            path_spec["responses"][str(code)] = response_spec

        method = "get"
        if transmute_func.creates:
            method = "put"
        elif transmute_func.updates:
            method = "post"
        elif transmute_func.deletes:
            method = "delete"
        return {method: path_spec}



class FlaskTransmuteException(Exception):
    pass


class ApiException(FlaskTransmuteException):
    """ raising this will return a "success": false with some details """

    def __init__(self, message=None, status_code=400):
        super(ApiException, self).__init__(message)
        self.status_code = status_code


class UnableToParseBody(ApiException):
    """ unable to parse body as a valid object. """


class UnrecognizedContentType(ApiException):
    """ unable to parse body as a valid object. """


class NotFoundException(ApiException):
    """ a convenience exception for 404 not found. """
    def __init__(self, message=None):
        super(NotFoundException, self).__init__(message, status_code=404)



#!/usr/bin/env python
import os
from setuptools import setup, find_packages

base = os.path.dirname(os.path.abspath(__file__))

README_PATH = os.path.join(base, "README.rst")

install_requires = [
    'Flask',
    # we're using restplus because it has a nicely bundled
    # set of statics for swagger.
    'flask-restplus',
    'tornado',
    'pyyaml',
]

tests_require = []

setup(name='flask-transmute',
      version='0.2.8',
      description='a flask plugin to generate routes from objects.',
      long_description=open(README_PATH).read(),
      author='Yusuke Tsutsumi',
      author_email='yusuke@tsutsumi.io',
      url='',
      packages=find_packages(),
      package_data={"flask_transmute": ["swagger-ui/*"]},
      install_requires=install_requires,
      classifiers=[
          'Development Status :: 3 - Alpha',
          'Operating System :: MacOS',
          'Operating System :: POSIX :: Linux',
          'Topic :: System :: Software Distribution',
          'License :: OSI Approved :: MIT License',
          'Programming Language :: Python',
          'Programming Language :: Python :: 2',
          'Programming Language :: Python :: 2.6',
          'Programming Language :: Python :: 2.7',
          'Programming Language :: Python :: 3',
          'Programming Language :: Python :: 3.2',
          'Programming Language :: Python :: 3.3',
          'Programming Language :: Python :: 3.4',
      ],
      tests_require=tests_require
)



"""
contains the logic that handles serialization to and from markup
types.
"""
from abc import ABCMeta, abstractmethod


class ContentTypeSerializer(object):
    __metaclass__ = ABCMeta

    @abstractmethod
    def foo():
        pass



from .utils import SWAGGER_TYPEMAP


class Definitions(object):
    """
    this class handler returning the correct swagger type spec.
    it will handler returning the proper schema value.
    """

    def __init__(self):
        self._element_name = "definitions"
        self._definitions = ModelDict()

    def get(self, model_or_cls):
        if isinstance(model_or_cls, dict):
            return self._get_reference(model_or_cls)

        if isinstance(model_or_cls, list):
            return {
                "type": "array",
                "items": self.get(model_or_cls[0]),
                "collectionFormat": "multi"
            }

        for typ, typ_name in SWAGGER_TYPEMAP.items():
            if issubclass(model_or_cls, typ):
                return {"type": typ_name}

        return self._get_reference(model_or_cls)

    def add_to_spec(self, spec):
        """ add definitions to the swagger spec """
        spec[self._element_name] = self._definitions

    def _get_reference(self, model_or_cls):
        self.add_model(model_or_cls)

        reference = "#/{0}/{1}".format(
            self._element_name,
            self._definitions.to_id(model_or_cls)
        )
        return {"$ref": reference}

    def add_model(self, model_or_cls):
        if not isinstance(model_or_cls, dict):
            model = model_or_cls.transmute_schema
        else:
            model = model_or_cls

        schema = self._expand_schema(model)
        self._definitions[model_or_cls] = schema
        return schema

    def _expand_schema(self, schema):
        properties = {}
        for property_name, details in schema["properties"].items():
            properties[property_name] = self.get(details["type"])

        return {
            "type": "object",
            "properties": properties,
            "required": schema["required"]
        }


class ModelDict(dict):
    """
    a dictionary that accepts class names or dictionaries
    as a key.
    """
    def __init__(self, *args, **kwargs):
        # for dictionaries, a stringified key
        # won't work in the swagger spec.
        # instead, we assign them ids instead.
        self._ids = {}

    def __getitem__(self, key, value):
        key = self.to_id(key)
        return super(ModelDict, self).__getitem__(key, value)

    def __setitem__(self, key, value):
        key = self.to_id(key)
        return super(ModelDict, self).__setitem__(key, value)

    def __contains__(self, key):
        key = self.to_id(key)
        return super(ModelDict, self).__contains__(key)

    def to_id(self, key):
        if isinstance(key, dict):
            ids_key = str(key)
            if ids_key not in self._ids:
                model_key = str(len(self._ids))
                self._ids[ids_key] = model_key
                return model_key
            else:
                return self._ids[ids_key]
        elif isinstance(key, type):
            key = "{0}.{1}".format(key.__module__, key.__name__)
        return key



from .exceptions import InvalidSchema


def validate_schema(schema):
    """ validate that the schema is a valid transmute schema """
    required_keys = ["properties"]

    errors = []
    for key in required_keys:
        if key not in schema:
            errors.append("key {0} is required for a valid schema.".format(key))

    properties = schema.get("properties", {})

    if not isinstance(properties, dict):
        errors.append("properties key must be a dict")
    else:
        for name, details in properties.items():
            if not isinstance(name, str):
                errors.append("expected property name to be a string, found {0}".format(str(type(name))))

            if not isinstance(details, dict):
                errors.append("expected property details to be a dict, found {0}".format(str(type(details))))
            else:
                errors.extend(_validate_schema_property(details))

    required = schema.get("required", [])

    if not isinstance(required, list):
        errors.append("expected 'required' to be a list. found {0}".format(str(type(required))))
    elif isinstance(properties, dict):
        for r in required:
            if r not in properties:
                errors.append("{0} is specified as a required parameter, but no definition exists.".format(r))
    if errors:
        raise InvalidSchema(errors, schema)


def _validate_schema_property(prop_dict):
    """ validate that the transmute schema property is valid """
    errors = []
    if "type" in prop_dict:
        if not _is_valid_type_definition(prop_dict["type"]):
            errors.append("{0} is not a valid type definition.".format(prop_dict["type"]))
    else:
        errors.append("property details requires parameter 'type'")
    return errors


def _is_valid_type_definition(type_definition):
    if isinstance(type_definition, list):
        if len(type_definition) != 1:
            return False
        return _is_valid_type_definition(type_definition[0])

    elif isinstance(type_definition, dict):
        for key, value in type_definition.items():
            if not isinstance(key, str):
                return False
            if not _is_valid_type_definition(value):
                return False
        return True
    return isinstance(type_definition, type)



from .exceptions import UnableToParseBody
import json
import yaml


def parse_body(data, content_type):
    """
    given a content type, parse the body appropriately. raises an
    unable to parse exception if content is not valid markup.

    returns (dict, mime_type)
    """
    if "yaml" in content_type:
        return yaml.load(data), "application/x-yaml"
    elif "json" in content_type:
        return json.loads, "application/json"


def dump_body(data, content_type):
    """
    given a content type, dumps the data to the desired type.
    desired type. An UnrecognizedContentType exception will be
    raised if an non-accepted content type is passed.
    """



import pkg_resources
import copy
from flask import jsonify, render_template
from flask_restplus.apidoc import apidoc
from .definitions import Definitions
from .paths import Paths

SWAGGER_FILES_ROOT = pkg_resources.resource_filename(
    "flask_transmute", "swagger-ui"
)

EXAMPLE_SWAGGER_JSON = {
    "info": {"title": "myApi", "version": "1.0"},
    "swagger": "2.0",
    "tags": [{
        "name": "pet"
    }],
    "paths": {
        "/deck/add_card": {
            "post": {
                "tags": ["card"],
                "summary": "add a card to a deck.",
                "description": "",
                "produces": ["application/json"],
                "parameters": [{
                    "in": "body",
                    "name": "body",
                    "required": True,
                    #"schema": {
                    #   "$ref": "#/definitions/Card"
                    #}
                    "schema": {
                        "type": "object",
                        "properties": {
                            "name": {"type": "string"},
                            "description": {"type": "string"}
                        }
                    }
                }],
                "responses": {
                    "200": {"description": "good input"}
                }
            }
        }
    },
    "definitions": {
        "Card": {
            "type": "object",
            "properties": {
                "name": {"type": "string"},
                "description": {"type": "string"}
            }
        }
    }
}


class Swagger(object):
    """
    A flask-transmute extension that generates
    a swagger route
    """
    swagger_version = "2.0"

    def __init__(self, title, version, route_prefix="/"):
        self._title = title
        self._version = version
        self._swagger_object = None
        self._definitions = Definitions()
        self._paths = Paths(self._definitions)
        self._restplus_apidoc = copy.deepcopy(apidoc)
        self._route_prefix = route_prefix

    def init_app(self, app):
        swagger_json = self._generate_swagger_json(app)
        swagger_route = self._route_prefix + "swagger.json"
        swagger_ui_route = self._route_prefix

        @app.route(swagger_route)
        def return_swagger_json():
            # return jsonify(EXAMPLE_SWAGGER_JSON)
            return jsonify(swagger_json)

        @self._restplus_apidoc.route(swagger_ui_route)
        def return_swagger_ui():
            return render_template("swagger-ui.html",
                                   specs_url=swagger_route)

        app.register_blueprint(self._restplus_apidoc)

    def _generate_swagger_json(self, app):
        """
        write a swagger json object. this method should be
        re-run if any changes are made to the routing.
        """
        self._paths.extract_from_app(app)

        swagger_object = {
            "swagger": self.swagger_version,
            "info": {
                "title": self._title,
                "version": self._version
            },
            "paths": {}
        }
        self._paths.add_to_spec(swagger_object)
        self._definitions.add_to_spec(swagger_object)

        return swagger_object
